==> audiochain-1.1.5-src.tgz: org/audiochain/devices/drum/DrumSchema.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Reader;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.sound.sampled.UnsupportedAudioFileException;

/**
 * This represents a very simple drum computer schema. The schema can be created
 * as a text file by the following principle.
 * 
 * This class represents and reads a file like this:
 * 
 * <pre>
 * # Definition of Clips. The keys must be unique. A clip is an instrument, normally a drum hit or a single tone with a
 * # defined beginning and a soon ending. The durance of a clip should not be much longer than one or two beats.
 * b=/lib/base.wav
 * s=/lib/snare.wav
 * c=/lib/cymbal.wav 0.75
 * 
 * # Definition of a Rhythm (consists of a name, sequence definitions and a rhythm sequence
 * 
 * # Rhytm Name
 * Rhytm One
 * 
 * # Sequence Definition (a unique name per sequence definition) and multiple patterns (divided by the pipe character)
 * # The Pattern 'b . s .' has four beats: each divided by a space. Each beat may have one or more clip keys (b, s or bs)
 * # The special clip key '.' means no clip at the corresponding beat. So a beat is a 'slot' or a 'tick' that can be filled with 
 * # a clip.
 * 01 | b . s . | b b s . |
 * 02 | b . b s | b b s . |
 * 03 | b b s . | b . s . |
 * 
 * # A rhythm sequence, by the above sequence definition names, divided by spaces 
 * 01 01 02 02 03 03 02 02 01 
 * 
 * 
 * # Next Rhythm
 * My Beat
 * 
 * 01 | b . . . s . . . | b . b . s . . . |
 * 02 | b . . b s . . . | b b . b s . . . |
 * 03 | b b b . s . . . | b . . b s . . . |
 * 
 * 02 01 01 02 03 02 03 
 * 
 * 
 * Funky Beat
 * 
 * 01 | b c . c s c . c | b bc b c s c . c |
 * 02 | b c b c s c . c | b bc . c s c . c |
 * 
 * 01 02 01 02 01 02 01 02
 * </pre>
 * 
 * @author Florian Paul
 * 
 */
public class DrumSchema {
	private final Collection<DrumClip> clips = new LinkedHashSet<DrumClip>();
	private final Collection<DrumRhythm> rhythms = new LinkedHashSet<DrumRhythm>();

	public Collection<DrumClip> getClips() {
		return clips;
	}

	public Collection<DrumRhythm> getRhythms() {
		return rhythms;
	}
	
	Collection<String> getDrumRhythmNames() {
		Collection<String> ret = new LinkedHashSet<String>();
		for (DrumRhythm drumRhythm : rhythms) {
			ret.add(drumRhythm.getName());
		}
		return ret;
	}

	public DrumRhythm getDrumRhythmByName(String name) {
		for (DrumRhythm ret : rhythms) {
			if (ret.getName().equals(name)) {
				return ret;
			}
		}
		throw new IllegalArgumentException("The drum rhythm '" + name + "' does not exist.");
	}

	public boolean hasDrumRhythm(String name) {
		for (DrumRhythm ret : rhythms) {
			if (ret.getName().equals(name)) {
				return true;
			}
		}
		return false;
	}

	public DrumClip getDrumClip(String name) {
		for (DrumClip clip : clips) {
			if (clip.getName().equals(name)) {
				return clip;
			}
		}
		return null;
	}
	
	void validate() throws DrumSchemaFormatException {
		if (clips.isEmpty()) {
			throw new DrumSchemaFormatException("The drum schema has no clips.");
		}
		if (rhythms.isEmpty()) {
			throw new DrumSchemaFormatException("The drum schema has no rhythms.");
		}
		for (DrumRhythm rhythm : rhythms) {
			rhythm.validate();
		}
	}

	@Override
	public String toString() {
		String linesep = System.getProperty("line.separator");
		StringBuilder sb = new StringBuilder();
		for (DrumClip clip : clips) {
			sb.append(clip.toString()).append(linesep);
		}
		sb.append(linesep);
		for (DrumRhythm rhythm : rhythms) {
			sb.append(rhythm.toString()).append(linesep).append(linesep).append(linesep);
		}
		return sb.toString();
	}

	public static DrumSchema readDrumSchema(Reader reader, File ... clipSearchPaths) throws IOException, DrumSchemaFormatException, UnsupportedAudioFileException {
		Pattern clipPattern = Pattern.compile("^\\s*(\\S+)\\s*=\\s*(\\S.*\\S)\\s*$");
		Pattern clipLevelPattern = Pattern.compile("^\\s*(\\S+)\\s*=\\s*(\\S.*\\S)\\s*(\\d[,\\.]\\d+)\\s*$");
		Pattern rhythmNamePattern = Pattern.compile("^\\s*(\\S.*\\S)\\s*$");
		Pattern rhythmPattern = Pattern.compile("^\\s*(\\d+[\\s\\d]*\\d+)\\s*$");
		Pattern sequencePattern = Pattern.compile("^\\s*(\\d+)\\s*\\|(.*)\\|\\s*$");

		DrumSchema ret = new DrumSchema();

		BufferedReader br = new BufferedReader(reader);
		String line;
		DrumRhythm rhythm = null;
		Integer beatsPerPattern = null;
		Integer pulse = null;

		while ((line = br.readLine()) != null) {
			line = line.trim();

			if (line.isEmpty()) {
				continue;
			}
			if (line.startsWith("#")) {
				continue;
			}
			Matcher m;
			if ((m = clipLevelPattern.matcher(line)).matches() || (m = clipPattern.matcher(line)).matches()) {
				String clipKey = m.group(1);
				if (ret.getDrumClip(clipKey) != null) {
					throw new DrumSchemaFormatException("Clip '" + clipKey + "' is already defined (see '" + line + "').");
				}
				String declaredClipFileNameOrPath = m.group(2);
				// at first assume an absolute path
				File clipFile = searchClipFile(declaredClipFileNameOrPath, clipSearchPaths);
				DrumClip clip;
				if(m.groupCount()==3) {
					String levelStr = m.group(3).replace(',', '.');
					float level = Float.parseFloat(levelStr);
					clip = new DrumClip(clipKey, clipFile, level);
				} else {
					clip = new DrumClip(clipKey, clipFile);
				}
				ret.getClips().add(clip);
			} else if ((m = sequencePattern.matcher(line)).matches()) {
				// 01 | b c . c s c . c | b bc b c s c . c |
				String sequenceName = m.group(1); // 01
				String patternDefinitions = m.group(2); // ' b c . c s c . c | b
														// bc b c s c . c '
				if (rhythm == null) {
					throw new DrumSchemaFormatException("No Rhythm Name defined for sequence '" + line + "'.");
				}
				DrumSequence seq = rhythm.getDrumSequenceByName(sequenceName);
				if (seq != null) {
					throw new DrumSchemaFormatException("Sequence '" + sequenceName + "' is already defined for rhythm '" + rhythm.getName() + "' (see '" + line + "').");
				}
				seq = new DrumSequence(sequenceName);
				rhythm.getSequenceDefinitions().add(seq);
				String[] patternDefs = patternDefinitions.split("\\|");
				for (String patternDef : patternDefs) { // ' b bc b c s c . c '
					DrumPattern pattern = new DrumPattern();
					seq.getPatterns().add(pattern);

					// collect non empty strings
					List<String> beatDefList = new LinkedList<String>();
					String[] beatDefs = patternDef.split(" ");
					for (String beatDef : beatDefs) { // bc
						if (!beatDef.trim().isEmpty()) {
							beatDefList.add(beatDef);
						}
					}

					// identify the pulse of the rhythm by the first occurence
					// of a pattern
					final int beatCount = beatDefList.size();
					if (pulse == null) {
						float calcPulse = beatCount;
						while (calcPulse > 4) {
							calcPulse = calcPulse * 0.5F;
						}
						if (((int) calcPulse) != calcPulse) {
							throw new DrumSchemaFormatException("The number of beats must be a multiple of two and additionally a multiple of either three or four, but " + beatCount
									+ " beats were found in pattern (see '" + line + "').");
						}
						pulse = (int) calcPulse;
						rhythm.setBeatsPerPulse(beatCount / pulse);
						beatsPerPattern = beatCount;
					}

					if (beatCount != beatsPerPattern) {
						throw new DrumSchemaFormatException("The number of beats per pattern must be " + beatsPerPattern + " (see '" + line + "').");
					}

					for (String beatDef : beatDefList) {
						DrumBeat beat = new DrumBeat();
						pattern.getBeats().add(beat);
						for (DrumClip clip : ret.getClips()) {
							String clipKey = clip.getName();
							if (beatDef.contains(clipKey)) {
								DrumHit hit = new DrumHit(clipKey); // b
								beat.getHits().add(hit);
							}
						}
					}
				}
			} else if ((m = rhythmPattern.matcher(line)).matches()) {
				String rhythmDef = m.group(1); // 01 01 02 02 03 03 02 02 01
				if (rhythm == null) {
					throw new DrumSchemaFormatException("No Rhythm Name defined for rhythm '" + line + "'.");
				}
				String[] sequenceNames = rhythmDef.split(" ");
				for (String sequenceName : sequenceNames) { // 01
					if (sequenceName.trim().isEmpty()) {
						continue;
					}
					DrumSequence seq = rhythm.getDrumSequenceByName(sequenceName);
					if (seq == null) {
						throw new DrumSchemaFormatException("No Sequence found with name '" + sequenceName + "' (see '" + line + "').");
					}
					rhythm.getSequences().add(seq);
				}
			} else if ((m = rhythmNamePattern.matcher(line)).matches()) {
				String name = m.group(1);
				rhythm = new DrumRhythm(name);
				pulse = null;
				ret.getRhythms().add(rhythm);
			}
		}
		
		br.close();

		ret.validate();
		
		return ret;
	}

	static File searchClipFile(String declaredClipFileNameOrPath, File... clipSearchPaths) throws FileNotFoundException {
		File clipFile = new File(declaredClipFileNameOrPath);
		String clipFileName = clipFile.getName();
		if (!clipFile.exists() || !clipFile.canRead()) {
			String filesep = System.getProperty("file.separator");
			// assume a relative path to the given dirs (inclusive possible subdirs)
			for(File clipSearchPath : clipSearchPaths) {
				String searchPath = clipSearchPath.getPath() + filesep;
				clipFile = new File(searchPath + declaredClipFileNameOrPath);
				if (clipFile.exists() && clipFile.canRead()) {
					break;
				}
				
				clipFile = new File(searchPath + clipFileName);
				if (clipFile.exists() && clipFile.canRead()) {
					break;
				}
			}
			if (!clipFile.exists() || !clipFile.canRead()) {
				throw new FileNotFoundException("Clip '" + declaredClipFileNameOrPath + "' can not be found.");
			}
		}
		return clipFile;
	}
}

==> audiochain-1.1.6-src.tgz: org/audiochain/devices/drum/DrumSchema.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.drum;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.sound.sampled.UnsupportedAudioFileException;

/**
 * This represents a very simple drum computer schema. The schema can be created
 * as a text file by the following principle.
 * 
 * This class represents and reads a file like this:
 * 
 * <pre>
 * # Definition of Clips. The keys must be unique. A clip is an instrument, normally a drum hit or a single tone with a
 * # defined beginning and a soon ending. The durance of a clip should not be much longer than one or two beats.
 * b=/lib/base.wav
 * s=/lib/snare.wav
 * c=/lib/cymbal.wav 0.75
 * 
 * # Definition of a Rhythm (consists of a name, sequence definitions and a rhythm sequence
 * 
 * # Rhytm Name
 * Rhytm One
 * 
 * # Sequence Definition (a unique name per sequence definition) and multiple patterns (divided by the pipe character)
 * # The Pattern 'b . s .' has four beats: each divided by a space. Each beat may have one or more clip keys (b, s or bs)
 * # The special clip key '.' means no clip at the corresponding beat. So a beat is a 'slot' or a 'tick' that can be filled with 
 * # a clip.
 * 01 | b . s . | b b s . |
 * 02 | b . b s | b b s . |
 * 03 | b b s . | b . s . |
 * 
 * # A rhythm sequence, by the above sequence definition names, divided by spaces 
 * 01 01 02 02 03 03 02 02 01 
 * 
 * 
 * # Next Rhythm
 * My Beat
 * 
 * 01 | b . . . s . . . | b . b . s . . . |
 * 02 | b . . b s . . . | b b . b s . . . |
 * 03 | b b b . s . . . | b . . b s . . . |
 * 
 * 02 01 01 02 03 02 03 
 * 
 * 
 * Funky Beat
 * 
 * 01 | b c . c s c . c | b bc b c s c . c |
 * 02 | b c b c s c . c | b bc . c s c . c |
 * 
 * 01 02 01 02 01 02 01 02
 * </pre>
 * 
 * @author Florian Paul
 * 
 */
public class DrumSchema {
	private final Collection<DrumClip> clips = new LinkedHashSet<DrumClip>();
	private final Collection<DrumRhythm> rhythms = new LinkedHashSet<DrumRhythm>();
	private long lastModified;

	public Collection<DrumClip> getClips() {
		return clips;
	}

	public Collection<DrumRhythm> getRhythms() {
		return rhythms;
	}
	
	Collection<String> getDrumRhythmNames() {
		Collection<String> ret = new LinkedHashSet<String>();
		for (DrumRhythm drumRhythm : rhythms) {
			ret.add(drumRhythm.getName());
		}
		return ret;
	}

	public DrumRhythm getDrumRhythmByName(String name) {
		for (DrumRhythm ret : rhythms) {
			if (ret.getName().equals(name)) {
				return ret;
			}
		}
		throw new IllegalArgumentException("The drum rhythm '" + name + "' does not exist.");
	}

	public boolean hasDrumRhythm(String name) {
		for (DrumRhythm ret : rhythms) {
			if (ret.getName().equals(name)) {
				return true;
			}
		}
		return false;
	}

	public DrumClip getDrumClip(String name) {
		for (DrumClip clip : clips) {
			if (clip.getName().equals(name)) {
				return clip;
			}
		}
		return null;
	}
	
	void validate() throws DrumSchemaFormatException {
		if (clips.isEmpty()) {
			throw new DrumSchemaFormatException("The drum schema has no clips.");
		}
		if (rhythms.isEmpty()) {
			throw new DrumSchemaFormatException("The drum schema has no rhythms.");
		}
		for (DrumRhythm rhythm : rhythms) {
			rhythm.validate();
		}
	}

	@Override
	public String toString() {
		String linesep = System.getProperty("line.separator");
		StringBuilder sb = new StringBuilder();
		for (DrumClip clip : clips) {
			sb.append(clip.toString()).append(linesep);
		}
		sb.append(linesep);
		for (DrumRhythm rhythm : rhythms) {
			sb.append(rhythm.toString()).append(linesep).append(linesep).append(linesep);
		}
		return sb.toString();
	}

	public static DrumSchema readDrumSchema(Reader reader, File ... clipSearchPaths) throws IOException, DrumSchemaFormatException, UnsupportedAudioFileException {
		Pattern clipPattern = Pattern.compile("^\\s*(\\S+)\\s*=\\s*(\\S.*\\S)\\s*$");
		Pattern clipLevelPattern = Pattern.compile("^\\s*(\\S+)\\s*=\\s*(\\S.*\\S)\\s*(\\d[,\\.]\\d+)\\s*$");
		Pattern sequencePattern = Pattern.compile("^\\s*(\\S+)\\s*\\|(.*)\\|\\s*$");

		DrumSchema ret = new DrumSchema();

		BufferedReader br = new BufferedReader(reader);
		String line;
		DrumRhythm rhythm = null;
		Integer beatsPerPattern = null;
		Integer pulse = null;

		while ((line = br.readLine()) != null) {
			line = line.trim();

			if (line.isEmpty()) {
				continue;
			}
			if (line.startsWith("#")) {
				continue;
			}
			Matcher m;
			if ((m = clipLevelPattern.matcher(line)).matches() || (m = clipPattern.matcher(line)).matches()) {
				String clipKey = m.group(1);
				if (ret.getDrumClip(clipKey) != null) {
					throw new DrumSchemaFormatException("Clip '" + clipKey + "' is already defined (see '" + line + "').");
				}
				String declaredClipFileNameOrPath = m.group(2);
				// at first assume an absolute path
				File clipFile = searchClipFile(declaredClipFileNameOrPath, clipSearchPaths);
				DrumClip clip;
				if(m.groupCount()==3) {
					String levelStr = m.group(3).replace(',', '.');
					float level = Float.parseFloat(levelStr);
					clip = new DrumClip(clipKey, clipFile, level);
				} else {
					clip = new DrumClip(clipKey, clipFile);
				}
				ret.getClips().add(clip);
			} else if ((m = sequencePattern.matcher(line)).matches()) {
				// 01 | b c . c s c . c | b bc b c s c . c |
				String sequenceName = m.group(1); // 01
				String patternDefinitions = m.group(2); // ' b c . c s c . c | b
														// bc b c s c . c '
				if (rhythm == null) {
					throw new DrumSchemaFormatException("No Rhythm Name defined for sequence '" + line + "'.");
				}
				DrumSequence seq = rhythm.getDrumSequenceByName(sequenceName);
				if (seq != null) {
					throw new DrumSchemaFormatException("Sequence '" + sequenceName + "' is already defined for rhythm '" + rhythm.getName() + "' (see '" + line + "').");
				}
				seq = new DrumSequence(sequenceName);
				rhythm.getSequenceDefinitions().add(seq);
				String[] patternDefs = patternDefinitions.split("\\|");
				for (String patternDef : patternDefs) { // ' b bc b c s c . c '
					DrumPattern pattern = new DrumPattern();
					seq.getPatterns().add(pattern);

					// collect non empty strings
					List<String> beatDefList = new LinkedList<String>();
					String[] beatDefs = patternDef.split(" ");
					for (String beatDef : beatDefs) { // bc
						if (!beatDef.trim().isEmpty()) {
							beatDefList.add(beatDef);
						}
					}

					// identify the pulse of the rhythm by the first occurence
					// of a pattern
					final int beatCount = beatDefList.size();
					if (pulse == null) {
						float calcPulse = beatCount;
						while (calcPulse > 4) {
							calcPulse = calcPulse * 0.5F;
						}
						if (((int) calcPulse) != calcPulse) {
							throw new DrumSchemaFormatException("The number of beats must be a multiple of two and additionally a multiple of either three or four, but " + beatCount
									+ " beats were found in pattern (see '" + line + "').");
						}
						pulse = (int) calcPulse;
						rhythm.setBeatsPerPulse(beatCount / pulse);
						beatsPerPattern = beatCount;
					}

					if (beatCount != beatsPerPattern) {
						throw new DrumSchemaFormatException("The number of beats per pattern must be " + beatsPerPattern + " (see '" + line + "').");
					}

					for (String beatDef : beatDefList) {
						DrumBeat beat = new DrumBeat();
						pattern.getBeats().add(beat);
						for (DrumClip clip : ret.getClips()) {
							String clipKey = clip.getName();
							if (beatDef.contains(clipKey)) {
								DrumHit hit = new DrumHit(clipKey); // b
								beat.getHits().add(hit);
							}
						}
					}
				}
			} else {
				Collection<DrumSequence> drumSequences = parseDrumSequences(line, rhythm);
				if (rhythm == null || drumSequences == null) {
					if (rhythm != null && rhythm.getSequences().isEmpty()) {
						throw new DrumSchemaFormatException("No Rhythm sequence found for rhythm '" + rhythm.getName() + "'.");
					}
					rhythm = new DrumRhythm(line);
					pulse = null;
					ret.getRhythms().add(rhythm);
				} else {
					if(drumSequences.isEmpty()) {
						throw new DrumSchemaFormatException("No Rhythm sequence found for rhythm '" + rhythm.getName() + "'.");
					} else {
						rhythm.getSequences().addAll(drumSequences);
					}
				}
			}
		}
		
		br.close();

		ret.validate();
		
		return ret;
	}
	
	private static Collection<DrumSequence> parseDrumSequences(String line, DrumRhythm rhythm) {
		if (line.trim().isEmpty() || rhythm == null) {
			return null;
		}
		String[] sequenceNames = line.split(" "); // 01 01 02 02 03 03 02 02 01
		Collection<DrumSequence> ret = new ArrayList<DrumSequence>();
		for (String sequenceName : sequenceNames) { // 01
			if (sequenceName.trim().isEmpty()) {
				continue;
			}
			DrumSequence seq = rhythm.getDrumSequenceByName(sequenceName);
			if (seq == null) {
				return null;
			}
			ret.add(seq);
		}
		return ret;
	}

	static File searchClipFile(String declaredClipFileNameOrPath, File... clipSearchPaths) throws FileNotFoundException {
		File clipFile = new File(declaredClipFileNameOrPath);
		String clipFileName = clipFile.getName();
		if (!clipFile.exists() || !clipFile.canRead()) {
			String filesep = System.getProperty("file.separator");
			// assume a relative path to the given dirs (inclusive possible subdirs)
			for(File clipSearchPath : clipSearchPaths) {
				String searchPath = clipSearchPath.getPath() + filesep;
				clipFile = new File(searchPath + declaredClipFileNameOrPath);
				if (clipFile.exists() && clipFile.canRead()) {
					break;
				}
				
				clipFile = new File(searchPath + clipFileName);
				if (clipFile.exists() && clipFile.canRead()) {
					break;
				}
			}
			if (!clipFile.exists() || !clipFile.canRead()) {
				throw new FileNotFoundException("Clip '" + declaredClipFileNameOrPath + "' can not be found.");
			}
		}
		return clipFile;
	}

	public long getLastModified() {
		return lastModified;
	}

	public void setLastModified(long lastModified) {
		this.lastModified = lastModified;
	}
}
